/* ***********************************************  */
/*   Projet  : Recherche de solution de l'Ane rouge */
/*   Date    : avril 2009                           */
/*   Auteur  : Climoo                               */
/*   Fichier : kl_definitions.h                     */
/*   Description : En tete contenant toutes         */
/*   les definitions des structures, constantes,    */
/*   et fonctions de bases du programme             */
/*                                                  */
/* **************************************************/

/*La largeur d'un plateau de jeu Klotski*/
#define KL_LARGEUR 4

/*La hauteur d'un plateau de jeu Klotski*/
#define KL_HAUTEUR 5

/*Le nombre de cases du plateau de jeu*/
#define KL_SURFACE (KL_HAUTEUR*KL_LARGEUR)

/*Le nombre de cases vides sur le jeu, qui reste constant d'un probleme Klotski a l'autre*/
#define KL_NB_CASES_VIDES 2

/*Constante caractere pour un caractere correspondant a l'ane rouge*/
#define KL_ANE_ROUGE 'L'

/*Constante pour le nombre de piece de taille 1 (petits carres)*/
#define KL_NB_PIECES_1 4

/*Constante chaine de caracteres pour l'ensemble des pieces de taille 1*/
#define KL_CARAC_PIECES_1 "CDEF" 

/*Constante pour le nombre de piece de taille 2 (rectangles verticaux ou horizontaux)*/
#define KL_NB_PIECES_2 5

/*Constante chaine de caracteres pour l'ensemble des pieces de taille 2*/
#define KL_CARAC_PIECES_2 "GHIJK"

/*Constante pour le nombre de symboles differents sur un plateau de jeu*/
#define KL_NB_SYMBOLES (KL_NB_PIECES_1 + KL_NB_PIECES_2 + KL_NB_CASES_VIDES + 1)

/*Constante correspondant au nombre de cases utilises par l'ane rouge (carre de 2*2)*/
#define KL_TAILLE_ANE 4

/*Tableau contant des indices des points sur lesquels doit se trouver l'ane rouge en fin de partie*/
#define KL_PTS_ANE_GAGNANT {13,14,17,18}

/*Definition d'une piece du jeu par un caractere*/
typedef unsigned char KL_Piece;

/*Definition d'une direction par un caractere*/
typedef unsigned char KL_Direction;

/*Constante pour le nombre de directions possibles*/
#define KL_NB_DIRECTIONS 4 

/*Constantes pour chacune des directions*/
#define HAUT 0
#define BAS 1
#define GAUCHE 2
#define DROITE 3


/*La structure correspondant au mouvement d'une piece precise sur le plateau a un moment donne*/
typedef struct{
  KL_Piece pie;    /*La piece concernee par le mouvement*/
  KL_Direction dir;/*La direction du mouvement: Haut, Bas, Gauche ou Droite*/
 
}KL_Mouvement;


/*La structure d'un plateau de jeu Klotski*/
typedef struct{
  KL_Piece surface[KL_SURFACE];/*La surface ou plateau de jeu. Chaque case correspond a un point*/
  KL_Mouvement mvt;            /*Dernier mouvement qui a conduit a cette situation*/
}KL_Situation;

/*La structure d'une solution de jeu Klotski*/
typedef struct{
  KL_Mouvement *mvts; /*Les mouvements ayant conduit a cette situation.
			Attention :Le mouvement d'indice 0 est le dernier mouvement*/
  int nbmvts;         /*Le nombre de mouvements conduisant a la solution.
			Vaut -1 s'il n'y a pas de solution.*/
} KL_Solution;

/*Accesseur lecture/ecriture pour une case du plateau de jeu, 
  etant donne une KL_Situation et un indice. Retourne une KL_PIECE.
  L'indice doit etre valable, pas de verification effectuee.*/
#define getCase(sur,i) ((sur).surface[(i)])

/*Macro testant si un caractere est celui correspondant a l'ane rouge dans un fichier Klotski*/
#define estCaracAneRouge(c) ((c)==KL_ANE_ROUGE)

/*Macro testant si un caractere est un caractere d'une piece de taille 1 du fichier klotski*/
#define estCaracPiece1(c) (((c) !=0) && (strchr(KL_CARAC_PIECES_1,(c))) != NULL) 
  
/*Macro testant si un caractere est un caractere d'une piece de taille 2 du fichier klotski*/
#define estCaracPiece2(c) (((c) !=0) && (strchr(KL_CARAC_PIECES_2,(c))) != NULL)

/*Macro testant si un caractere est un caractere d'une piece quelconque du fichier klotski*/
#define estCaracPiece(c) (estCaracPiece1((c)) || estCaracPiece2((c)) || ((c) == KL_ANE_ROUGE))

/*Macro permettant de tester si un caractere est un caractere correspondant au vide ou non*/
#define estCaracVide(c) ((c)=='A' || (c)=='B')

/*Macro permettant d'obtenir la taille (nombre de cases) d'une piece du jeu*/
#define getTaillePiece(p) (estCaracAneRouge(p)? : KL_TAILLE_ANE : ((esCaracPiece2(p))? 2 : 1))

/*Accesseur pour obtenir l'indice d'une case se trouvant au dessus, au dessous, a gauche ou a droite,
  de celle de l'indice donne.
  Si celui n'existe pas, on retourne -1*/
#define getIndDessus(ind) ( ((ind) >= KL_LARGEUR)? ((ind) - KL_LARGEUR) : -1 )

#define getIndDessous(ind) ( ((ind) < (KL_SURFACE-KL_LARGEUR))? ((ind) + KL_LARGEUR) : -1 )

#define getIndGauche(ind) ( ((ind) % KL_LARGEUR != 0)? ((ind) - 1) : -1 )

#define getIndDroite(ind) ( ((ind) % KL_LARGEUR != (KL_LARGEUR-1))? ((ind) + 1) : -1 )
