/* ***********************************************  */
/*   Projet  : Recherche de solution de l'Ane rouge */
/*   Date    : avril 2009                           */
/*   Auteur  : Climoo                               */
/*   Fichier : kl_deplacements.c                    */
/*   Description : Module charge du                 */
/*   deplacement des pieces du jeu                  */
/*                                                  */
/* **************************************************/

#include <string.h>
#include <stdlib.h>
#include "allocations.h"
#include "kl_definitions.h"
#include "kl_tabSituations.h"
#include "kl_deplacements.h"

/*Decale vers la direction dir une partie d'une piece a partir de l'indice numCase.
Exemples :
 |--| 
 |A |
 |A |
 |--|
Deplacer la piece A, en case 0 (en haut a gauche) vers la droite donnera :
 |--| 
 | A|
 |A |
 |--|
Il faut rappeler la fonction sur la case 2 (en bas a gauche) pour obtenir
le deplacement complet de la piece A.

Aucune verification n'est effectue. Le deplacement est suppose valide.*/
static void deplacerCase(KL_Situation *kl, int numCase, KL_Direction dir);

/*Fonction permettant la recopie de la situation klsrc dans la situation kldest.
  kldest est allouee dynamiquement.*/
static void recopierSituation(KL_Situation **kldest, KL_Situation *klsrc);

/*Macro permettant de savoir si une piece a deja ete deplacee auparavant dans une direction donnee*/
#define aEteDeplacee(drapeaux,p,dir) ((drapeaux)[KL_NB_DIRECTIONS*(dir)+((p)-'A')] == 1)

/*Macro definissant une piece comme deplacee dans une direction donnee */
#define setDeplacee(drapeaux,p,dir) ((drapeaux)[KL_NB_DIRECTIONS*(dir)+((p)-'A')] = 1)

/*Fonctions permettant de verifier que la piece a l'indice iv de la situation Klotski pere
  peut etre deplacee dans une direction.
  Le tableau (drapeaux) est mis a jour en consequence.
  Les situations connus sont aussi mises a jour si le deplacement a lieu.
  La situation resultante est placee dans nv (mis a NULL si pas de deplacement possible).
  On retourne 1 si la piece a effectivement ete deplacee, 0 sinon.*/
static char tenterDeplacementBas(KL_Situations connus,KL_Situation **nv,KL_Situation *pere, int iv,char *depls);
static char tenterDeplacementHaut(KL_Situations connus,KL_Situation **nv,KL_Situation *pere, int iv,char *depls);
static char tenterDeplacementGauche(KL_Situations connus,KL_Situation **nv,KL_Situation *pere, int iv,char *depls);
static char tenterDeplacementDroite(KL_Situations connus,KL_Situation **nv,KL_Situation *pere, int iv,char *depls);

static void deplacerCase(KL_Situation *kl, int numCase, KL_Direction dir){
  int indice1 = 0; /*indice qui  va contenir la piece apres le deplacement */
  int indice2 = numCase; /*indice qui va contenir le vide apres le deplacement*/
  KL_Piece pie = getCase(*kl,numCase);
  KL_Piece vide = 0;
  int ok;
  switch(dir){
  case BAS:
    indice1 = numCase + KL_LARGEUR;/*case au dessous*/
    ok = getIndDessus(indice2);
    vide = getCase(*kl,getIndDessous(numCase));
    while ( ok != -1 && getCase(*kl,ok) == pie) {
      indice2 -= KL_LARGEUR ;
      ok = getIndDessus(indice2);
    }
    break;
  case HAUT:
    indice1 = numCase - KL_LARGEUR;/*case au dessus*/
    ok = getIndDessous(indice2);
    vide = getCase(*kl,getIndDessus(numCase));
    while (ok !=-1 &&  getCase(*kl,ok) == pie) {
      indice2 += KL_LARGEUR ;
      ok = getIndDessous(indice2);
    }
    break;
  case GAUCHE:
    indice1 = numCase - 1; /*case a gauche*/
    ok = getIndDroite(indice2);
    vide = getCase(*kl,getIndGauche(numCase));
    while ( ok !=-1 && getCase(*kl,ok) == pie) {
      indice2 ++ ;
      ok = getIndDroite(indice2);
    }
    break;
  case DROITE: 
    indice1 = numCase + 1; /*case a droite*/
    ok = getIndGauche(indice2);
    vide = getCase(*kl,getIndDroite(numCase));
    while ( ok !=-1 && getCase(*kl,ok) == pie) {
      indice2 -- ;
      ok = getIndGauche(indice2);
    }
    break;
  }
  getCase(*kl, indice1) = pie ;
  getCase(*kl, indice2) = vide;
  return ;
}

static void recopierSituation(KL_Situation **kldest, KL_Situation *klsrc){
  *kldest = (KL_Situation*)allouer(sizeof(KL_Situation));
  memcpy((*kldest)->surface,klsrc->surface,KL_SURFACE*sizeof(KL_Piece));
}

static char tenterDeplacementBas(KL_Situations connus,KL_Situation **nv,KL_Situation *pere, int iv,char *depls){
  KL_Piece p;
  int iv1, iv2;
  *nv = NULL;
  p = getCase(*pere,iv); 
  if ((!estCaracVide(p)) && (!aEteDeplacee(depls,p,BAS))){
    iv1 = getIndGauche(iv);
    iv2 = getIndDroite(iv);
    /*Si la piece est de largeur 1*/
    if  ( ((iv1 == -1) || (getCase(*pere,iv1) != p))  && ((iv2 == -1) || (getCase(*pere,iv2) != p)) ) {
      recopierSituation(nv,pere);
      deplacerCase(*nv,iv,BAS);
      setDeplacee(depls,p,BAS);
    }
    /*Si la piece est de largeur 2, l'autre case contenant la piece etant a gauche*/
    else if ((iv1 != -1) && (getCase(*pere,iv1) == p) && (estCaracVide(getCase(*pere,getIndDessous(iv1))))) {
      recopierSituation(nv,pere);
      deplacerCase(*nv,iv,BAS);
      deplacerCase(*nv,iv1,BAS);
      setDeplacee(depls,p,BAS);
    }
    /*Si la piece est de largeur 2, l'autre case contenant la piece etant a droite*/
    else if ((iv2 != -1) && (getCase(*pere,iv2) == p) && (estCaracVide(getCase(*pere,getIndDessous(iv2))))){ 
      recopierSituation(nv,pere);
      deplacerCase(*nv,iv,BAS);
      deplacerCase(*nv,iv2,BAS);
      setDeplacee(depls,p,BAS);
    }
    if ((*nv != NULL)){ /*S'il y a effectivement un deplacement possible*/
      if (ajouterSituationConnue(connus,*nv)){ /*Si on a trouve une nouvelle configuration, jusqu'alors inconnue*/
	(*nv)->mvt.pie = p;
	(*nv)->mvt.dir = BAS;
	return 1;
      }
      else /*Si la situation existait deja*/
	free(*nv);
    }
  }
  return 0;
}

char tenterDeplacementHaut(KL_Situations connus,KL_Situation **nv,KL_Situation *pere, int iv,char *depls){
  KL_Piece p;
  int iv1, iv2;
  p = getCase(*pere,iv);          
  if ((!estCaracVide(p)) && (!aEteDeplacee(depls,p,HAUT))){
    iv1 = getIndGauche(iv);
    iv2 = getIndDroite(iv);
    if  ( ((iv1 == -1) || (getCase(*pere,iv1) != p))  && ((iv2 == -1) || (getCase(*pere,iv2) != p)) ) {
      recopierSituation(nv,pere);
      deplacerCase(*nv,iv,HAUT);
      setDeplacee(depls,p,HAUT);
    }
    else if ((iv1 != -1) && (getCase(*pere,iv1) == p) && (estCaracVide(getCase(*pere,getIndDessus(iv1))))) {
      recopierSituation(nv,pere);
      deplacerCase(*nv,iv,HAUT);
      deplacerCase(*nv,iv1,HAUT);
      setDeplacee(depls,p,HAUT);
    }
    else if ((iv2 != -1) && (getCase(*pere,iv2) == p) && (estCaracVide(getCase(*pere,getIndDessus(iv2))))){ 
      recopierSituation(nv,pere);
      deplacerCase(*nv,iv,HAUT);
      deplacerCase(*nv,iv2,HAUT);
      setDeplacee(depls,p,HAUT);
    }
    if (aEteDeplacee(depls,p,HAUT)) {
      if (ajouterSituationConnue(connus,*nv)){ /*Si on a trouve une nouvelle configuration, jusqu'alors inconnue*/
	(*nv)->mvt.pie = p;
	(*nv)->mvt.dir = HAUT;
	return 1;
      }
      else
	free(*nv);
    }
  }
  return 0;
}

char tenterDeplacementGauche(KL_Situations connus,KL_Situation **nv,KL_Situation *pere, int iv, char *depls){
  KL_Piece p;
  int iv1, iv2;
  p = getCase(*pere,iv);          
  if (((!estCaracVide(p))) && (!aEteDeplacee(depls,p,GAUCHE))){
    iv1 = getIndDessus(iv);
    iv2 = getIndDessous(iv);
    if  ( ((iv1 == -1) || (getCase(*pere,iv1) != p))  && ((iv2 == -1) || (getCase(*pere,iv2) != p)) ) {
      recopierSituation(nv,pere);
      deplacerCase(*nv,iv,GAUCHE);
      setDeplacee(depls,p,GAUCHE);
    }
    else if ((iv1 != -1) && (getCase(*pere,iv1) == p) && (estCaracVide(getCase(*pere,getIndGauche(iv1))))) {
      recopierSituation(nv,pere);
      deplacerCase(*nv,iv,GAUCHE);
      deplacerCase(*nv,iv1,GAUCHE);
      setDeplacee(depls,p,GAUCHE);
    }
    else if ((iv2 != -1) && (getCase(*pere,iv2) == p) && (estCaracVide(getCase(*pere,getIndGauche(iv2))))){ 
      recopierSituation(nv,pere);
      deplacerCase(*nv,iv,GAUCHE);
      deplacerCase(*nv,iv2,GAUCHE);
      setDeplacee(depls,p,GAUCHE);
    }
    if (aEteDeplacee(depls,p,GAUCHE)) {
      if(ajouterSituationConnue(connus,*nv)){ /*Si on a trouve une nouvelle configuration, jusqu'alors inconnue*/
	(*nv)->mvt.pie = p;
	(*nv)->mvt.dir = GAUCHE;
	return 1;
      }
      else
	free(*nv);
    }
  }
  return 0;
}

static char tenterDeplacementDroite(KL_Situations connus,KL_Situation **nv,KL_Situation *pere, int iv, char *depls){
  KL_Piece p;
  int iv1, iv2;
  p = getCase(*pere,iv);          
  if (((!estCaracVide(p))) && (!aEteDeplacee(depls,p,DROITE))){
    iv1 = getIndDessus(iv);
    iv2 = getIndDessous(iv);
    if  ( ((iv1 == -1) || (getCase(*pere,iv1) != p))  && ((iv2 == -1) || (getCase(*pere,iv2) != p)) ) {
      recopierSituation(nv,pere);
      deplacerCase(*nv,iv,DROITE);
      setDeplacee(depls,p,DROITE);
    }
    else if ((iv1 != -1) && (getCase(*pere,iv1) == p) && (estCaracVide(getCase(*pere,getIndDroite(iv1))))) {
      recopierSituation(nv,pere);
      deplacerCase(*nv,iv,DROITE);
      deplacerCase(*nv,iv1,DROITE);
      setDeplacee(depls,p,DROITE);
    }
    else if ((iv2 != -1) && (getCase(*pere,iv2) == p) && (estCaracVide(getCase(*pere,getIndDroite(iv2))))){ 
      recopierSituation(nv,pere);
      deplacerCase(*nv,iv,DROITE);
      deplacerCase(*nv,iv2,DROITE);
      setDeplacee(depls,p,DROITE);
    }
    if (aEteDeplacee(depls,p,DROITE)) {
      if (ajouterSituationConnue(connus,*nv)){ /*Si on a trouve une nouvelle configuration, jusqu'alors inconnue*/
	(*nv)->mvt.pie = p;
	(*nv)->mvt.dir = DROITE;
	return 1;
      }	
      else
	free(*nv);
    }    
  }
  return 0;
}

int trouverDeplacements(KL_Situations connus, KL_Situation *pere, KL_Situation **lesfils){
  int i = KL_SURFACE -1, nbmvt = 0;
  int cptv = 0 ; /*Indice qui va compter le nombre de cases vides trouvees sur le plateau*/
  int iv ;
  char depls[KL_NB_SYMBOLES*KL_NB_DIRECTIONS];/*Nombre de pieces*nombre de directions*/
  memset(depls,0,KL_NB_SYMBOLES*KL_NB_DIRECTIONS);
  /*Recherche des cases vides */
  while(cptv < KL_NB_CASES_VIDES){
    if (estCaracVide(getCase(*pere,i))){    /*Si on a trouve une case vide*/
      iv = getIndDessus(i); 
      if (iv != -1)  /*s'il y a une case au dessus de la case vide*/
	nbmvt += tenterDeplacementBas(connus,&lesfils[nbmvt],pere,iv,depls);

      iv = getIndDessous(i);
      if (iv != -1)  /*s'il y a une case au dessous de la case vide*/
	nbmvt += tenterDeplacementHaut(connus,&lesfils[nbmvt],pere,iv,depls);

      iv = getIndGauche(i);
      if (iv != -1)  /*s'il y a une case a gauche de la case vide*/
	nbmvt += tenterDeplacementDroite(connus,&lesfils[nbmvt],pere,iv,depls);
      
      iv = getIndDroite(i);
      if (iv != -1)  /*s'il y a une case a droite de la case vide*/
	nbmvt += tenterDeplacementGauche(connus,&lesfils[nbmvt],pere,iv,depls);

      cptv++;
    }
    i--;
  }
  return nbmvt ;
}
