/* ***********************************************  */
/*   Projet  : Recherche de solution de l'Ane rouge */
/*   Date    : avril 2009                           */
/*   Auteur  : Climoo                               */
/*   Fichier : kl_main.c                            */
/*   Description : Module contenant la fonction     */
/*   principale du programme                        */
/*                                                  */
/* **************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "kl_definitions.h"
#include "kl_lectureFichier.h"
#include "kl_resolution.h"
#include "kl_affichage.h"

#define VERSION "1.0"

#define OPT_AIDE "-h"
#define OPT_VERS "-v"
#define OPT_FIC  "-f"


static void afficherAide(char *nomPrgm){
  printf("%s : usage :\n",nomPrgm);
  printf(" [%s]         : affiche les options possibles\n",OPT_AIDE);
  printf(" [%s]         : affiche la version du programme\n",OPT_VERS);
  printf(" [%s fichier] : effectue la resolution du fichier\n",OPT_FIC);
}

static void afficherVersion(char *nomPrgm){
  printf("%s : version %s\n",nomPrgm, VERSION);
  printf("Avril 2009\n");
  printf("Les sources presentees sont libre de droits\n");
}

static char *verifierArgs(int argc, char **argv){
  int i = 1;
  char *nomProg = argv[0];
  char *nomFic = NULL;
  if (argc <= 1){
    afficherAide(argv[0]);
    return NULL;
  }
  while (i < argc){
    if (strcmp(argv[i],OPT_AIDE) == 0){
      afficherAide(nomProg);
      return NULL;
    }
    else if (strcmp(argv[i],OPT_VERS) == 0){
      afficherVersion(nomProg);
      return NULL;
    }
    else if (strcmp(argv[i],OPT_FIC) == 0){
      if (i+1 >= argc){
	afficherAide(nomProg);
	return NULL;
      }
      if (nomFic != NULL){
	fprintf(stderr,"%s [%s fichier] : un seul fichier en parametre\n",nomProg,OPT_FIC);
	return NULL;
      }
      else{
	nomFic = argv[i+1];
	i+=2;
      }
    }
    else{
      afficherAide(nomProg);
      return NULL;
    }
  }
  return nomFic;
}

int main(int argc, char **argv){
  char *nomFic;
  KL_Situation *depart = NULL;
  KL_Solution solution ;
  nomFic = verifierArgs(argc, argv);
  if (nomFic == NULL) /*Si pas de resolution*/
    return 0;
  depart = lireFichier(nomFic);
  if (depart != NULL){
    afficherPlateau(depart);
    resoudre(depart,&solution);
    afficherSolution(&solution);
    free(solution.mvts);
  }
  else {
    fprintf(stderr,"%s : Echec de lecture - Fin du programme\n", nomFic);
    return 1;
  }
  return 0;
}
