/* ***********************************************  */
/*   Projet  : Recherche de solution de l'Ane rouge */
/*   Date    : avril 2009                           */
/*   Auteur  : Climoo                               */
/*   Fichier : kl_tabSituations.c                   */
/*   Description : Module charge l'ajout de         */
/*   nouvelles situations (ou configurations)       */  
/*   dans l'ensemble des situations deja connues    */       
/*                                                  */
/* **************************************************/

#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "allocations.h"
#include "kl_definitions.h"
#include "kl_tabSituations.h"

/*Constante pour le nombre de points utilises pour calculer la cle de hachage*/
#define KL_NB_PTS_IMPT 6

/*Tableau constant des points utilises pour calculer la cle de hachage*/
#define KL_PTS_IMPT {1,4,8,9,12,17}

/*Constante pour la taille par defaut de la table de hachage.*/
#define TAILLE_HASH_TABLE 15000

/*Fonction permettant l'insertion dans une liste chainee
  d'une nouvelle situation Klotski.*/
static void insererSituationListe(Cellule **cel, KL_Situation* kl);

/*Constante pour le nombre de symboles de remplacement*/
#define KL_NB_REMP 5

/*Symbole remplacant une piece de taille 1 (petit carre)*/
#define REMP_P1        0

/*Symbole remplacant une piece de taille 2 horizontale*/
#define REMP_P2_H      1

/*Symbole remplacant une piece de taille 2 verticale*/
#define REMP_P2_V      2

/*Symbole remplacant l'ane rouge*/
#define REMP_ANE_ROUGE 3

/*Symbole remplacant une case vide*/
#define REMP_VIDE      4 

/*Fonction qui initialise la table de conversions*/
static void initialiserTabConversion(KL_Conv tconv, KL_Situation *kl);

/*Macro qui renvoie le symbole de conversion d'une piece, etant une table de conversion*/
#define getSigneConv(tab,pie) ((tab)[pie - 'A'])

/*Fonction permettant de connaitre la cle de hachage d'une situation
  Klotski. On retourne la valeur de cette cle.*/
static int cleHash(KL_Conv *tconv,KL_Situation *kl);

/*Fonction permettant de decider si kl1 est equivalent a kl2.
  Retourne 1 si oui, 0 sinon.*/
static char estEquivalent(KL_Conv *tconv,KL_Situation *kl1,KL_Situation *kl2);

/*Macro calculant l'indice symetrique par rapport 
 a l'axe vertical du jeu. L'indice en parametre doit etre compris
 entre 0 et KL_SURFACE.*/
#define symetrique(ind) ((KL_LARGEUR - 1 -((ind) % (KL_LARGEUR)))+ (((ind) / KL_LARGEUR) * KL_LARGEUR))

/*Macro renvoyant le minimum entre deux valeurs*/
#define min(x,y) (((x) < (y))? (x) : (y))

static void insererSituationListe(Cellule **cel, KL_Situation* kl){
  Cellule * tmp= (Cellule *)allouer(sizeof(Cellule)) ;
  tmp->sit = kl;
  tmp->suivant = (*cel);
  *cel = tmp;  /*L'insertion est faite en tete de la liste*/
}


/*
Case :
de 0 a 1  : equivalent vide
de 2 a 5  : equivalent piece de taille 1 (C, D, E, F)
de 6 a 10 : equivalent piece de taille 2 (G, H, I, J, K)
11        : equivalent ane rouge
*/
static void initialiserTabConversion(KL_Conv tconv, KL_Situation *kl){
  KL_Piece pieces2[] = KL_CARAC_PIECES_2, *p;
  int i = 0;
  int j;
  p = pieces2; /*p <- premier caractere d'une piece de taille 2*/
  for (i = 0 ; i < KL_NB_CASES_VIDES ; i++)             /*Pour les cases vides*/
    tconv[i] = REMP_VIDE; 
  for (; i < (KL_NB_PIECES_1 + KL_NB_CASES_VIDES) ; i++) /*Pour les pieces de taille 1*/
    tconv[i] = REMP_P1; 
  for (; i < (KL_NB_PIECES_2 + KL_NB_PIECES_1 + KL_NB_CASES_VIDES); i++){ /*Pour les pieces de taille 2*/
    j = 0; /*Determiner s'il s'agit de piece horizontales ou verticales*/
    while (getCase(*kl,j) != *p)
      j++;
    if ((getCase(*kl,getIndDessous(j)) == *p) || (getCase(*kl,getIndDessus(j)) == *p))
      tconv[i] = REMP_P2_V;
    else
      tconv[i] = REMP_P2_H;
    p++;
  }
  tconv[i] = REMP_ANE_ROUGE; /*Remplacement de l'ane rouge*/
}

/* Voir documentation pour comprendre la fonction de cle de hachage */
static int cleHash(KL_Conv *tconv,KL_Situation *kl){
  int pts[]= KL_PTS_IMPT;
  int i = 0;
  int res = 0;   /*Resultat contenant la cle de hachage*/
  int puiss = 1;
  char conv1, conv2;
  for (i = 0; i< KL_NB_PTS_IMPT ; i++){
    conv1 = getSigneConv(*tconv,getCase(*kl,pts[i]));
    conv2 = getSigneConv(*tconv,getCase(*kl,symetrique(pts[i])));
    res += min(conv1,conv2) * puiss;
    puiss *= KL_NB_REMP;
  }  
  return (res % TAILLE_HASH_TABLE);/*Assurer une valeur entre 0 et TAILLE_HASH_TABLE*/
}


/*Initialise le tableau des situations connues*/
void initialiserTabSituations(KL_Situations *tab, KL_Situation *kl){
  initialiserTabConversion((tab->tconv), kl);
  tab->cels = (Cellule **)allouer0(sizeof(Cellule*)*TAILLE_HASH_TABLE);
}

static char estEquivalent(KL_Conv *tconv,KL_Situation *kl1,KL_Situation *kl2){
  int i = KL_SURFACE - 1 ;
  char sym = 1, ident = 1;
  char conv1, conv2;
  while ((i >= 0) && (sym || ident)){ 
    conv1 = getSigneConv(*tconv,getCase(*kl1,i));
    conv2 = getSigneConv(*tconv,getCase(*kl2,i));
    ident = (ident && (conv1 == conv2)); /*Test de d'identicite*/
    conv2 = getSigneConv(*tconv,getCase(*kl2,symetrique(i)));
    sym = (sym && (conv1 == conv2));     /*Test de symetrie*/
    i--;
  }
  return (sym || ident); /*Renvoie 1 si la situation est similaire*/
}

char ajouterSituationConnue(KL_Situations tab,KL_Situation *kl){
  int cle = cleHash(&(tab.tconv),kl);
  Cellule *cell= tab.cels[cle];
  char trouveEqui = 0;
  /*Regarder chaque situation se trouvant dans la liste a l'indice donne par la cle*/
  /*La situation que l'on veut ajouter est potentiellement equivalente a l'un de celle ci.*/
  while ((cell != NULL) && (!trouveEqui)){
    trouveEqui = estEquivalent(&(tab.tconv),kl,cell->sit);
    cell = cell->suivant;
  }
  if (trouveEqui)
    return 0;
  insererSituationListe(&(tab.cels[cle]),kl);/*Insertion seulement si on ne connaissait pas deja la situation*/
  return 1;
}


void libererTabSituations(KL_Situations tab){
  int i = 0;
  Cellule *cell,*tmp;
  while (i < TAILLE_HASH_TABLE){
    cell = tab.cels[i];
    while (cell != NULL){
      tmp = cell->suivant;
      free(cell->sit); /*Liberation de la situation Klotski*/
      free(cell);      /*Liberation de la cellule*/
      cell = tmp;
    }
    i++;
  }
  free(tab.cels); /*Liberation du tableau de cellules*/
}
