/* ***********************************************  */
/*   Projet  : Recherche de solution de l'Ane rouge */
/*   Date    : avril 2009                           */
/*   Auteur  : Climoo                               */
/*   Fichier : kl_tabSituations.h                   */
/*   Description : En-tete du Module charge l'ajout */
/*   de nouvelles situations (ou configurations)    */  
/*   dans l'ensemble des situations deja connues    */       
/*                                                  */
/* **************************************************/

/*Structure d'une cellule d'une liste chainee.*/
typedef struct cel{
  KL_Situation *sit;   /*La situation connue, element de la cellule*/
  struct cel *suivant; /*Pointeur vers la cellule suivante*/
}Cellule;

/*Structure du tableau de conversion des pieces.
  A chaque piece, on va associer un nouveau symbole.
  Par exemple, toutes les pieces de taille 2 vertical auront le meme symbole associe.
  Ceci permet de comparer rapidement deux situations, pour savoir si elles sont
  equivalentes ou symetriques.*/
typedef char KL_Conv[KL_NB_SYMBOLES];

/*Structure des situation de Klotski deja connues*/
typedef struct {
  KL_Conv tconv; /*Tableau de conversion associee a la situation de Klotski de depart*/
  Cellule **cels;/*Table de hachage*/
}KL_Situations; 

/*Fonction permettant d'initialiser la table de hachage a vide.
  Doit etre appelle avant la premiere utilisation de la table.
  La table de conversion est initialisee egalement*/
void initialiserTabSituations(KL_Situations *tab, KL_Situation *kl);

/*Fonction permettant d'ajouter une situation Klotski a la table
 de hachage. Si une situation similaire est deja connue, elle
 ne sera pas ajoutee. Retourne 1 si l'insertion a eu lieu, 
 0 sinon.
 Remarque importante :
 Les situations symetriques sont considerees comme similaires.*/
char ajouterSituationConnue(KL_Situations tab,KL_Situation *kl);

/*Fonction permettant de liberer la memoire utilisee par la table
  de hachage. 
  Remarque importante :
  Les situations klotski memorisees sont aussi supprimees.*/
void libererTabSituations(KL_Situations tab);
